import pytest
from embedbase.settings import get_settings_from_file
from embedbase.embedding.openai import OpenAI
from embedbase.embedding.cohere import Cohere
from embedbase.embedding.base import Embedder



d = [
    "Bob is a human",
    "The quick brown fox jumps over the lazy dog",
]


@pytest.mark.asyncio
async def test_can_embed():
    settings = get_settings_from_file()
    embedders: Embedder = [
        OpenAI(settings.openai_api_key, settings.openai_organization),
        # Cohere(settings.cohere_api_key)
    ]
    for embedder in embedders:
        embeddings = await embedder.embed(d)
        assert len(embeddings) == 2
        assert len(embeddings[0]) == embedder.dimensions
        assert len(embeddings[1]) == embedder.dimensions




import logging
from typing import Optional

from embedbase.settings import Settings
from logging import Logger

def get_logger(settings: Optional[Settings]) -> Logger:
    logger = logging.getLogger("embedbase")
    logger.setLevel(settings.log_level or "INFO")
    handler = logging.StreamHandler()
    handler.setLevel(settings.log_level or "INFO")
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger



import asyncio
from typing import Coroutine, List, Optional
from pandas import DataFrame, Series
from embedbase.database import VectorDatabase
from embedbase.utils import BatchGenerator


class Supabase(VectorDatabase):
    def __init__(
        self,
        url: str,
        key: str,
    ):
        """
        :param url: supabase url
        :param key: supabase key
        """
        try:
            from supabase import create_client, Client

            self.supabase: Client = create_client(url, key)
            self.functions = self.supabase.functions()

        except ImportError:
            raise ImportError("Please install supabase with `pip install supabase`")

    async def select(
        self,
        ids: List[str] = [],
        hashes: List[str] = [],
        dataset_id: Optional[str] = None,
        user_id: Optional[str] = None,
    ) -> List[dict]:
        # either ids or hashes must be provided
        assert ids or hashes, "ids or hashes must be provided"

        req = self.supabase.table("documents").select("*")
        if ids:
            req = req.in_("id", ids)
        if hashes:
            req = req.in_("hash", hashes)
        if dataset_id:
            req = req.eq("dataset_id", dataset_id)
        if user_id:
            req = req.eq("user_id", user_id)
        return req.execute().data

    async def update(
        self,
        df: DataFrame,
        dataset_id: str,
        user_id: Optional[str] = None,
        batch_size: Optional[int] = 100,
        store_data: bool = True,
    ) -> Coroutine:
        df_batcher = BatchGenerator(batch_size)
        batches = [batch_df for batch_df in df_batcher(df)]

        async def _insert(batch_df: DataFrame):
            def _d(row: Series):
                data = {
                    "id": row.id,
                    "embedding": row.embedding,
                    "hash": row.hash,
                    "dataset_id": dataset_id,
                    "user_id": user_id,
                    "metadata": row.metadata,
                }
                if store_data:
                    data["data"] = row.data
                return data

            response = (
                self.supabase.table("documents")
                .upsert([_d(row) for _, row in batch_df.iterrows()])
                .execute()
            )
            return response

        # TODO: not sure truly parallel, python garbage
        results = await asyncio.gather(*[_insert(batch_df) for batch_df in batches])
        return results

    async def delete(
        self,
        ids: List[str],
        dataset_id: str,
        user_id: Optional[str] = None,
    ) -> None:
        req = self.supabase.table("documents").delete().eq("dataset_id", dataset_id)
        if user_id:
            req = req.eq("user_id", user_id)
        return req.in_("id", ids).execute()

    async def search(
        self,
        vector: List[float],
        top_k: Optional[int],
        dataset_ids: List[str],
        user_id: Optional[str] = None,
    ) -> List[dict]:
        d = {
            "query_embedding": vector,
            "similarity_threshold": 0.1,  # TODO: make this configurable
            "match_count": top_k,
            "query_dataset_ids": dataset_ids,
        }
        if user_id:
            d["query_user_id"] = user_id
        return (
            self.supabase.rpc(
                "match_documents",
                d,
            )
            .execute()
            .data
        )

    async def clear(self, dataset_id: str, user_id: Optional[str] = None) -> None:
        req = self.supabase.table("documents").delete().eq("dataset_id", dataset_id)
        if user_id:
            req = req.eq("user_id", user_id)
        return req.execute()

    async def get_datasets(self, user_id: Optional[str] = None) -> List[dict]:
        req = self.supabase.table("distinct_datasets").select("dataset_id", "documents_count")
        if user_id:
            req = req.eq("user_id", user_id)
        data = req.execute().data
        return data



"""
Tests at the end-to-end abstraction level.
"""

import math
from random import randint
from typing import List

import numpy as np
import pandas as pd
import pytest
from httpx import AsyncClient

from embedbase import get_app
from embedbase.database.base import VectorDatabase
from embedbase.database.postgres_db import Postgres
from embedbase.database.supabase_db import Supabase
from embedbase.embedding.openai import OpenAI
from embedbase.settings import get_settings_from_file
from tests.test_utils import unit_testing_dataset

vector_databases: List[VectorDatabase] = []


# before running any test initialize the databases
@pytest.fixture(scope="session", autouse=True)
def init_databases():
    settings = get_settings_from_file()

    vector_databases.append(Postgres())
    vector_databases.append(
        Supabase(
            url=settings.supabase_url,
            key=settings.supabase_key,
        )
    )


async def run_around_tests():
    for vector_database in vector_databases:
        await vector_database.clear(unit_testing_dataset)
        settings = get_settings_from_file()
        app = (
            get_app(settings)
            .use_db(vector_database)
            .use_embedder(OpenAI(settings.openai_api_key))
            .run()
        )
        yield app


@pytest.mark.asyncio
async def test_clear():
    async for app in run_around_tests():
        df = pd.DataFrame(
            [
                "".join(
                    [
                        chr(math.floor(97 + 26 * np.random.rand()))
                        for _ in range(randint(500, 800))
                    ]
                )
                for _ in range(10)
            ],
            columns=["text"],
        )
        async with AsyncClient(app=app, base_url="http://localhost:8000") as client:
            response = await client.post(
                f"/v1/{unit_testing_dataset}",
                json={
                    "documents": [
                        {
                            "data": text,
                            "metadata": {"somethibng": f"test_{i}"},
                        }
                        for i, text in enumerate(df.text.tolist())
                    ],
                },
            )
            assert response.status_code == 200
            json_response = response.json()
            assert len(json_response.get("results")) == 10

        async with AsyncClient(app=app, base_url="http://localhost:8000") as client:
            response = await client.post(
                f"/v1/{unit_testing_dataset}/search", json={"query": "bob", "top_k": 10}
            )
            assert response.status_code == 200
            json_response = response.json()
            assert json_response.get("query", "") == "bob"
            assert len(json_response.get("similarities")) == 10

        async with AsyncClient(app=app, base_url="http://localhost:8000") as client:
            response = await client.get(f"/v1/{unit_testing_dataset}/clear")
            assert response.status_code == 200
        # search now
        async with AsyncClient(app=app, base_url="http://localhost:8000") as client:
            response = await client.post(
                f"/v1/{unit_testing_dataset}/search", json={"query": "bob"}
            )
            assert response.status_code == 200
            json_response = response.json()
            assert json_response.get("query", "") == "bob"
            assert len(json_response.get("similarities")) == 0


@pytest.mark.asyncio
async def test_refresh_small_documents():
    async for app in run_around_tests():
        df = pd.DataFrame(
            [
                "".join(
                    [
                        chr(math.floor(97 + 26 * np.random.rand()))
                        for _ in range(randint(500, 800))
                    ]
                )
                for _ in range(10)
            ],
            columns=["text"],
        )
        async with AsyncClient(app=app, base_url="http://localhost:8000") as client:
            response = await client.post(
                f"/v1/{unit_testing_dataset}",
                json={
                    "documents": [
                        {
                            "data": text,
                        }
                        for i, text in enumerate(df.text.tolist())
                    ],
                },
            )
            assert response.status_code == 200
            json_response = response.json()
            assert len(json_response.get("results")) == 10


@pytest.mark.asyncio
async def test_sync_no_id_collision():
    async for app in run_around_tests():
        df = pd.DataFrame(
            ["foo" for _ in range(10)],
            columns=["text"],
        )
        async with AsyncClient(app=app, base_url="http://localhost:8000") as client:
            response = await client.post(
                f"/v1/{unit_testing_dataset}",
                json={
                    "documents": [
                        {
                            "data": text,
                        }
                        for i, text in enumerate(df.text.tolist())
                    ],
                },
            )
            assert response.status_code == 200
            json_response = response.json()
            # make sure all ids are unique
            ids = list(set([e["id"] for e in json_response.get("results")]))
            assert len(ids) == 10


@pytest.mark.asyncio
async def test_save_clear_data():
    async for app in run_around_tests():
        df = pd.DataFrame(
            ["bob is a human"],
            columns=["text"],
        )
        async with AsyncClient(app=app, base_url="http://localhost:8000") as client:
            response = await client.post(
                "/v1/unit_test",
                json={
                    "documents": [
                        {
                            "data": text,
                        }
                        for i, text in enumerate(df.text.tolist())
                    ],
                    "store_data": False,
                },
            )
            assert response.status_code == 200
            json_response = response.json()
            assert len(json_response.get("results")) == 1
        # now search shouldn't have the "data" field in the response
        async with AsyncClient(app=app, base_url="http://localhost:8000") as client:
            response = await client.post(
                "/v1/unit_test/search",
                json={"query": "bob"},
            )
            assert response.status_code == 200
            json_response = response.json()
            assert len(json_response.get("similarities")) > 0
            assert json_response.get("similarities")[0].get("data") is None


@pytest.mark.asyncio
async def test_health_properly_forward_headers():
    async for app in run_around_tests():
        import requests_mock

        # mock http://0.0.0.0:8000/v1/test
        with requests_mock.Mocker(
            real_http=True,
            case_sensitive=True,
        ) as m:
            m.post("http://0.0.0.0:8080/v1/test")
            async with AsyncClient(app=app, base_url="http://localhost:8000") as client:
                response = await client.get(
                    "/health",
                    headers={"Authorization": "Bearer 123"},
                )
                # TODO: any way to listen to the request and check the headers?
                # without using pcap or hacks like that lol?
                assert response.status_code == 200


@pytest.mark.asyncio
async def test_adding_twice_the_same_data_is_ignored():
    async for app in run_around_tests():
        d = [
            "The lion is the king of the jungle",
            "The lion is a large cat",
            "The lion is a carnivore",
        ]
        df = pd.DataFrame({"data": d})

        async def _i(results_length):
            async with AsyncClient(app=app, base_url="http://localhost:8000") as client:
                response = await client.post(
                    f"/v1/{unit_testing_dataset}",
                    json={
                        "documents": [
                            {
                                "data": data,
                            }
                            for i, data in enumerate(df.data.tolist())
                        ],
                    },
                )
                assert response.status_code == 200
                json_response = response.json()
                assert len(json_response.get("results")) == results_length

        # insert twice the same ting
        await _i(3)
        # should have been ignored but still return 3 to client
        await _i(3)

        # search should not have duplicates
        async with AsyncClient(app=app, base_url="http://localhost:8000") as client:
            response = await client.post(
                f"/v1/{unit_testing_dataset}/search", json={"query": "Feline animal"}
            )
            assert response.status_code == 200
            json_response = response.json()
            assert json_response.get("query", "") == "Feline animal"
            assert len(json_response.get("similarities")) == 3
            # check that there are no duplicates
            for idx, lion in enumerate(
                sorted([e["data"] for e in json_response.get("similarities")])
            ):
                assert lion == sorted(d)[idx], f"{lion} != {sorted(d)[idx]}"


@pytest.mark.asyncio
async def test_insert_large_documents_should_fail():
    async for app in run_around_tests():
        # large texts > 10.000 characters
        d = ["".join("agi " * 10_000) for _ in range(10)]
        df = pd.DataFrame({"data": d})

        async with AsyncClient(app=app, base_url="http://localhost:8000") as client:
            response = await client.post(
                f"/v1/{unit_testing_dataset}",
                json={
                    "documents": [
                        {
                            "data": data,
                        }
                        for i, data in enumerate(df.data.tolist())
                    ],
                },
            )
            assert response.status_code == 400

        # now search
        async with AsyncClient(app=app, base_url="http://localhost:8000") as client:
            response = await client.post(
                f"/v1/{unit_testing_dataset}/search", json={"query": "a", "top_k": 100}
            )
            assert response.status_code == 200
            json_response = response.json()
            assert json_response.get("query", "") == "a"
            assert len(json_response.get("similarities")) == 0


@pytest.mark.asyncio
async def test_get_datasets_without_auth():
    async for app in run_around_tests():
        """
        should create a dataset by inserting some data
        and return a list of datasets
        """
        async with AsyncClient(app=app, base_url="http://localhost:8000") as client:
            response = await client.get(
                f"/v1/datasets",
            )
            assert response.status_code == 200
            json_response = response.json()
            # shouldn't have "unit_testing_dataset" in the list
            ds_id = [e["dataset_id"] for e in json_response.get("datasets")]
            assert unit_testing_dataset not in ds_id

        d = [
            "The lion is the king of the jungle",
            "The lion is a large cat",
            "The lion is a carnivore",
        ]
        df = pd.DataFrame({"data": d})
        async with AsyncClient(app=app, base_url="http://localhost:8000") as client:
            response = await client.post(
                f"/v1/{unit_testing_dataset}",
                json={
                    "documents": [
                        {
                            "data": data,
                        }
                        for i, data in enumerate(df.data.tolist())
                    ],
                },
            )
            assert response.status_code == 200
            json_response = response.json()
            assert len(json_response.get("results")) == 3

        # get datasets
        async with AsyncClient(app=app, base_url="http://localhost:8000") as client:
            response = await client.get(
                f"/v1/datasets",
            )
            assert response.status_code == 200
            json_response = response.json()
            # should have "unit_testing_dataset" in the list
            ds_id = [e["dataset_id"] for e in json_response.get("datasets")]
            assert unit_testing_dataset in ds_id
            ds = [
                e
                for e in json_response.get("datasets")
                if e["dataset_id"] == unit_testing_dataset
            ][0]
            assert ds["documents_count"] == 3


@pytest.mark.asyncio
async def test_get_datasets_with_auth():
    async for app in run_around_tests():
        """
        an authenticated client
        should create a dataset by inserting some data
        and return a list of datasets
        another authenticated client
        should create a dataset by inserting some data
        and get a list of dataset, only his own
        """
        settings = get_settings_from_file()

        async def add_uid(request, call_next):
            request.scope["uid"] = "test"
            response = await call_next(request)
            return response

        app = (
            get_app(settings)
            .use_middleware(add_uid)
            .use_db(Postgres())
            .use_embedder(OpenAI(settings.openai_api_key))
        ).run()

        async with AsyncClient(app=app, base_url="http://localhost:8000") as client:
            response = await client.get(
                f"/v1/datasets",
            )
            assert response.status_code == 200
            json_response = response.json()
            # TODO: iterate on this test by insert first docs with this user then
            # check datasets
            assert json_response.get("datasets") == []


@pytest.mark.asyncio
async def test_cross_search_on_multiple_datasets():
    dataset_is = [
        f"{unit_testing_dataset}_lions",
        f"{unit_testing_dataset}_giraffes",
    ]
    async for app in run_around_tests():
        # clear all datasets
        for e in dataset_is:
            async with AsyncClient(app=app, base_url="http://localhost:8000") as client:
                response = await client.get(
                    f"/v1/{e}/clear",
                )
                assert response.status_code == 200
        lions = [
            "The lion is the king of the jungle",
            "The lion is a large cat",
            "The lion is a carnivore",
        ]
        giraffes = [
            "The giraffe is the tallest animal",
            "The giraffe is a large animal",
            "The giraffe is a herbivore",
        ]
        df_lions = pd.DataFrame({"data": lions})
        df_giraffes = pd.DataFrame({"data": giraffes})

        async with AsyncClient(app=app, base_url="http://localhost:8000") as client:
            response = await client.post(
                f"/v1/{unit_testing_dataset}_lions",
                json={
                    "documents": [
                        {
                            "data": data,
                        }
                        for i, data in enumerate(df_lions.data.tolist())
                    ],
                },
            )
            assert response.status_code == 200
            json_response = response.json()
            assert len(json_response.get("results")) == 3
            response = await client.post(
                f"/v1/{unit_testing_dataset}_giraffes",
                json={
                    "documents": [
                        {
                            "data": data,
                        }
                        for i, data in enumerate(df_giraffes.data.tolist())
                    ],
                },
            )
            assert response.status_code == 200
            json_response = response.json()
            assert len(json_response.get("results")) == 3

        async with AsyncClient(app=app, base_url="http://localhost:8000") as client:
            response = await client.post(
                f"/exo/search",
                json={
                    "query": "Lions and giraffes are animals that are found in Africa",
                    "top_k": 4,
                    "dataset_ids": dataset_is
                },
            )
            assert response.status_code == 200
            json_response = response.json()
            assert len(json_response.get("similarities")) == 4
            # words "lion" and "giraffe" should be found
            sentences = "\n".join(
                [e["data"] for e in json_response.get("similarities")]
            )
            assert "lion" in sentences
            assert "giraffe" in sentences



from typing import Tuple
import warnings
from fastapi import FastAPI, Request
from fastapi.responses import JSONResponse


def enable_firebase_auth(app: FastAPI):
    from firebase_admin import auth

    @app.middleware("http")
    async def firebase_auth(request: Request, call_next) -> Tuple[str, str]:
        # extract token from header
        for name, value in request.headers.items():  # type: bytes, bytes
            print(name, value)
            if name == "authorization":
                authorization = value
                break
        else:
            authorization = None

        if not authorization:
            return JSONResponse(
                status_code=401, content={"error": "missing authorization header"}
            )

        s = authorization.split(" ")

        if len(s) != 2:
            return JSONResponse(
                status_code=401, content={"error": "invalid authorization header"}
            )

        token_type, token = s
        assert (
            token_type == "Bearer"
        ), "Authorization header must be `Bearer` type. Like: `Bearer LONG_JWT`"

        try:
            token = token.strip()
            decoded_token = auth.verify_id_token(token)
            # add uid to scope
            request.scope["uid"] = decoded_token["uid"]
        except Exception as err:
            warnings.warning(f"Error verifying token: {err}")
            return JSONResponse(status_code=401, content={"error": "invalid token"})

        response = await call_next(request)
        return response



import json
from typing import List, Optional

from pandas import DataFrame, Series

from embedbase.database import VectorDatabase


class Postgres(VectorDatabase):
    def __init__(self, conn_str="postgresql://postgres:localdb@0.0.0.0/embedbase"):
        """
        Implements a vector database using postgres
        """
        try:
            import psycopg
            from pgvector.psycopg import register_vector

            self.conn = psycopg.connect(conn_str, dbname="embedbase")
            self.conn.autocommit = True
            self.conn.execute("CREATE EXTENSION IF NOT EXISTS vector")
            register_vector(self.conn)
            self.conn.execute(
                """
create table documents (
    id text primary key,
    data text,
    embedding vector (1536),
    hash text,
    dataset_id text,
    user_id text,
    metadata json,
    created_date TIMESTAMPTZ NOT NULL DEFAULT NOW()
);"""
            )
            self.conn.execute(
                """
create index on documents
using ivfflat (embedding vector_cosine_ops)
with (lists = 100);
"""
            )
            self.conn.execute(
                """
create or replace function match_documents (
  query_embedding vector(1536),
  similarity_threshold float,
  match_count int,
  query_dataset_ids text[],
  query_user_id text default null
)
returns table (
  id text,
  data text,
  score float,
  hash text,
  embedding vector(1536),
  metadata json
)
language plpgsql
as $$
begin
  return query
  select
    documents.id,
    documents.data,
    (1 - (documents.embedding <=> query_embedding)) as similarity,
    documents.hash,
    documents.embedding,
    documents.metadata
  from documents
  where 1 - (documents.embedding <=> query_embedding) > similarity_threshold
    and documents.dataset_id = any(query_dataset_ids)
    and (query_user_id is null or query_user_id = documents.user_id)
  order by documents.embedding <=> query_embedding
  limit match_count;
end;
$$;"""
            )
            self.conn.execute(
                """
CREATE OR REPLACE VIEW distinct_datasets AS
SELECT dataset_id, user_id, COUNT(*) AS documents_count
FROM documents
GROUP BY dataset_id, user_id;
"""
            )

        except ImportError:
            raise ImportError(
                "Please install pgvector and psycopg with `pip install pgvector psycopg`"
            )
        except psycopg.OperationalError:
            raise psycopg.OperationalError(
                "Please install postgresql and create a database named embedbase"
            )
        except Exception as e:
            print(e)

    async def select(
        self,
        ids: List[str] = [],
        hashes: List[str] = [],
        dataset_id: Optional[str] = None,
        user_id: Optional[str] = None,
    ) -> List[dict]:
        # either ids or hashes must be provided
        assert ids or hashes, "ids or hashes must be provided"
        from psycopg import sql

        query = """
select id, data, embedding, hash, metadata
from documents
where
    {conditions}
"""
        conditions = []
        if ids:
            conditions.append(
                sql.SQL("id in ({})").format(sql.SQL(",").join(map(sql.Literal, ids)))
            )
        if hashes:
            conditions.append(
                sql.SQL("hash in ({})").format(
                    sql.SQL(",").join(map(sql.Literal, hashes))
                )
            )
        if dataset_id:
            conditions.append(
                sql.SQL("dataset_id = {}").format(sql.Literal(dataset_id))
            )
        if user_id:
            conditions.append(sql.SQL("user_id = {}").format(sql.Literal(user_id)))
        data = []
        results = self.conn.execute(
            sql.SQL(query).format(conditions=sql.SQL(" and ").join(conditions))
        )
        for row in results:
            data.append(
                {
                    "id": row[0],
                    "data": row[1],
                    "embedding": row[2].tolist(),
                    "hash": row[3],
                    "metadata": row[4],
                }
            )
        return data

    async def update(
        self,
        df: DataFrame,
        dataset_id: str,
        user_id: Optional[str] = None,
        batch_size: Optional[int] = 100,
        store_data: bool = True,
    ):
        if len(df) == 0:
            return

        def _d(row: Series):
            data = [
                row.id,
                row.data if store_data else None,
                row.embedding,
                row.hash,
                dataset_id,
                user_id,
                json.dumps(row.metadata),
            ]
            return data

        values = [tuple(_d(row)) for _, row in df.iterrows()]
        num_columns = len(values[0])
        placeholders = ", ".join(
            ["(" + ", ".join(["%s"] * num_columns) + ")"] * len(values)
        )
        flat_values = [item for sublist in values for item in sublist]
        q = f"""
            INSERT INTO documents(id, data, embedding, hash, dataset_id, user_id, metadata)
            VALUES {placeholders}
            ON CONFLICT (id) DO UPDATE SET
                data = excluded.data,
                embedding = excluded.embedding,
                hash = excluded.hash,
                dataset_id = excluded.dataset_id,
                user_id = excluded.user_id,
                metadata = excluded.metadata
        """

        with self.conn.cursor() as cur:
            cur.execute(q, flat_values)

    async def delete(
        self,
        ids: List[str],
        dataset_id: str,
        user_id: Optional[str] = None,
    ) -> None:
        req = "delete from documents where id in %s and dataset_id = %s", (
            tuple(ids),
            dataset_id,
        )
        if user_id:
            req += f" and user_id = {user_id}"
        return [dict(row) for row in self.conn.execute(req)]

    async def search(
        self,
        vector: List[float],
        top_k: Optional[int],
        dataset_ids: List[str],
        user_id: Optional[str] = None,
    ) -> List[dict]:
        d = {
            "query_embedding": str(vector),
            "similarity_threshold": 0.1,  # TODO: make this configurable
            "match_count": top_k,
            "query_dataset_ids": dataset_ids,
            "query_user_id": user_id,
        }
        q = "select * from match_documents(%(query_embedding)s, %(similarity_threshold)s, %(match_count)s, %(query_dataset_ids)s, %(query_user_id)s)"
        results = self.conn.execute(q, d)
        if results.rowcount == 0:
            return []
        data = []
        for row in results:
            # tuple to dict
            data.append(
                {
                    "id": row[0],
                    "data": row[1],
                    "score": row[2],
                    "hash": row[3],
                    "embedding": row[4].tolist(),
                    "metadata": row[5],
                }
            )
        return data

    async def clear(self, dataset_id: str, user_id: Optional[str] = None) -> None:
        req = f"delete from documents where dataset_id = '{dataset_id}'"
        if user_id:
            req += f" and user_id = {user_id}"
        from psycopg import sql

        self.conn.execute(req)

    async def get_datasets(self, user_id: Optional[str] = None) -> List[dict]:
        req = "select * from distinct_datasets"
        if user_id:
            req += f" where user_id = '{user_id}'"
        results = self.conn.execute(req)
        if results.rowcount == 0:
            return []
        data = []
        for row in results:
            data.append(
                {
                    "dataset_id": row[0],
                    "documents_count": row[2],
                }
            )
        return data



from typing import Tuple
import warnings
from fastapi import FastAPI, Request
from fastapi.responses import JSONResponse


def enable_supabase_auth(app: FastAPI):
    @app.middleware("http")
    async def supabase_auth(request: Request, call_next) -> Tuple[str, str]:
        raise NotImplementedError


# an enum to pick from either pinecone, weaviate, or supabase
import asyncio
import itertools
from typing import List, Optional

from embedbase.database.base import VectorDatabase

# TODO: move this to VectorDatabase
async def batch_select(
    vector_database: VectorDatabase,
    hashes_to_fetch: List[str],
    dataset_id: Optional[str] = None,
    user_id: Optional[str] = None,
):
    """
    :param vector_database: vector database
    :param hashes_to_fetch: list of hashes
    :param dataset_id: dataset id
    :param user_id: user id
    """
    # this depend on the vector database used
    # supabase cannot deal with 200 for example
    n = 100
    hashes_to_fetch = [
        hashes_to_fetch[i : i + n] for i in range(0, len(hashes_to_fetch), n)
    ]

    async def _fetch(hashes) -> List[dict]:
        try:
            return await vector_database.select(
                hashes=hashes, dataset_id=dataset_id, user_id=user_id
            )
        except Exception as e:
            raise e

    existing_documents = await asyncio.gather(*[_fetch(hashes) for hashes in hashes_to_fetch])
    return itertools.chain.from_iterable(existing_documents)




import typing

from tenacity import (
    retry,
    retry_if_not_exception_type,
    stop_after_attempt,
    wait_exponential,
)

from embedbase.embedding.base import Embedder

try:
    import openai
except:
    pass


@retry(
    wait=wait_exponential(multiplier=1, min=1, max=3),
    stop=stop_after_attempt(3),
    # TODO: send pr/issue on https://github.com/openai/openai-python/blob/94428401b4f71596e4a1331102a6beee9d8f0bc4/openai/__init__.py#L25
    # To expose openai.AuthenticationError
    retry=retry_if_not_exception_type(openai.InvalidRequestError),
)
def embed_retry(
    input: typing.List[str],
) -> typing.List[dict]:
    """
    Embed a list of sentences and retry on failure
    :param input: list of sentences to embed
    :param provider: which provider to use
    :return: list of embeddings
    """
    return [
        e["embedding"]
        for e in openai.Embedding.create(input=input, model="text-embedding-ada-002")[
            "data"
        ]
    ]


class OpenAI(Embedder):
    """
    OpenAI Embedder
    """

    EMBEDDING_MODEL = "text-embedding-ada-002"
    EMBEDDING_CTX_LENGTH = 8191
    EMBEDDING_ENCODING = "cl100k_base"

    def __init__(
        self, openai_api_key: str, openai_organization: typing.Optional[str] = None
    ):
        super().__init__()
        try:
            import openai
            import tiktoken
        except ImportError:
            raise ImportError(
                "OpenAI is not installed. Install it with `pip install openai tiktoken`"
            )

        self.encoding = tiktoken.get_encoding(self.EMBEDDING_ENCODING)
        openai.api_key = openai_api_key
        openai.organization = openai_organization

    @property
    def dimensions(self) -> int:
        return 1536

    def is_too_big(self, text: str) -> bool:
        tokens = self.encoding.encode(text)
        if len(tokens) > self.EMBEDDING_CTX_LENGTH:
            return True

        return False

    async def embed(self, input: typing.List[str]) -> typing.List[typing.List[float]]:
        """
        Embed a list of texts
        :param texts: list of texts
        :return: list of embeddings
        """
        return embed_retry(input)



import re
from typing import Tuple, List


# unused
def is_upper_case_adjacent(s):
    s = s.split(" ")
    s = list(filter(lambda x: not re.match(r"[^A-Za-zäöüÄÖÜß]", x), s))
    for i in range(len(s) - 1):
        if s[i].isupper() and s[i + 1].isupper():
            return True
    return False


def string_similarity(
    str1: str, str2: str, substring_length: int = 2, case_sensitive: bool = False
) -> float:
    """
    Calculate similarity between two strings using
    https://en.wikipedia.org/wiki/S%C3%B8rensen%E2%80%93Dice_coefficient
    Computing time O(n)
    :param str1: First string to match
    :param str2: Second string to match
    :param substring_length: Optional. Length of substring to be used in calculating similarity. Default 2.
    :param case_sensitive: Optional. Whether you want to consider case in string matching. Default false;
    :return: Number between 0 and 1, with 0 being a low match score.
    """
    if not case_sensitive:
        str1 = str1.lower()
        str2 = str2.lower()

    if len(str1) < substring_length or len(str2) < substring_length:
        return 0

    m = {}
    for i in range(len(str1) - (substring_length - 1)):
        substr1 = str1[i : substring_length + i]
        m[substr1] = m.get(substr1, 0) + 1

    match = 0
    for j in range(len(str2) - (substring_length - 1)):
        substr2 = str2[j : substring_length + j]
        count = m.get(substr2, 0)

        if count > 0:
            match += 1
            m[substr2] = count - 1

    return (match * 2) / (len(str1) + len(str2) - ((substring_length - 1) * 2))


def group_by_similarity(
    sentences: List[str], threshold: float = 0.75
) -> List[List[str]]:
    """
    This is a function that takes a list of sentences
    and create groups of similarity and return them
    Complexity:
    Compute: O(n^2)
    :param sentences: The list of sentences to check.
    :type sentences: List[str]
    :param threshold: The threshold to consider similarity. Default 0.75.
    :type threshold: float, optional
    :return: Clusters of sentences.
    """
    groups = []
    # TODO: should ignore empty strings or option to ignore len < X
    for i in range(len(sentences)):
        sentence = sentences[i]
        is_in_group = False
        for j in range(len(groups)):
            group = groups[j]
            if len(group) == 0:
                continue
            similarity = sum([string_similarity(cur, sentence) for cur in group]) / len(
                group
            )
            if similarity >= threshold:
                is_in_group = True
                group.append(sentence)
        if not is_in_group:
            groups.append([sentence])
    return groups


def group_by_similarity_distinct(sentences: List[str], threshold: float = 0.75) -> map:
    """
    Given a list of sentences, group them by similarity,
    filter groups of size 2 or more,
    flatten the list by taking the shortest sentence in the groups
    :param sentences:
    :param threshold:
    :return:
    """
    groups = group_by_similarity(sentences, threshold)
    f = filter(lambda group: len(group) > 1, groups)
    m = map(lambda group: min(group, key=len), f)
    return m



from enum import Enum
from functools import lru_cache
import typing
import os
from pydantic_yaml import YamlModel


class VectorDatabaseEnum(str, Enum):
    pinecone = "pinecone"
    supabase = "supabase"
    weaviate = "weaviate"
    postgres = "postgres"


# an enum to pick from openai or cohere
class EmbeddingProvider(str, Enum):
    OPENAI = "openai"
    COHERE = "cohere"


class Settings(YamlModel):
    openai_api_key: typing.Optional[str] = None
    openai_organization: typing.Optional[str] = None
    supabase_url: typing.Optional[str] = None
    supabase_key: typing.Optional[str] = None
    
    log_level: str = "INFO"
    auth: typing.Optional[str] = None
    sentry: typing.Optional[str] = None
    firebase_service_account_path: typing.Optional[str] = None

@lru_cache()
def get_settings_from_file(path: str = "config.yaml"):
    """
    Read settings from a file, only supports yaml for now
    """
    settings = Settings.parse_file(path)

    # TODO: move
    # if firebase, init firebase
    if settings.auth and settings.auth == "firebase":
        import firebase_admin
        from firebase_admin import credentials

        cred = credentials.Certificate(settings.firebase_service_account_path)
        firebase_admin.initialize_app(cred)
    return settings



from abc import ABC, abstractmethod
from typing import Coroutine, List, Optional
from pandas import DataFrame


class VectorDatabase(ABC):
    """
    Base class for all vector databases
    """

    @abstractmethod
    async def select(
        self,
        ids: List[str] = [],
        hashes: List[str] = [],
        dataset_id: Optional[str] = None,
        user_id: Optional[str] = None,
    ) -> List[dict]:
        """
        :param ids: list of ids
        :param hashes: list of hashes
        :param dataset_id: dataset id
        :param user_id: user id
        :return: list of documents
        """
        raise NotImplementedError

    @abstractmethod
    async def update(
        self,
        df: DataFrame,
        dataset_id: str,
        user_id: Optional[str] = None,
        batch_size: Optional[int] = 100,
        store_data: bool = True,
    ) -> Coroutine:
        """
        :param df: dataframe
        :param dataset_id: dataset id
        :param user_id: user id
        :param batch_size: batch size
        :param store_data: store data in database?
        """
        raise NotImplementedError

    @abstractmethod
    async def delete(
        self, ids: List[str], dataset_id: str, user_id: Optional[str] = None
    ) -> None:
        """
        :param ids: list of ids
        :param dataset_id: dataset id
        :param user_id: user id
        """
        raise NotImplementedError

    @abstractmethod
    async def search(
        self,
        vector: List[float],
        top_k: Optional[int],
        dataset_ids: List[str],
        user_id: Optional[str] = None,
    ) -> List[dict]:
        """
        :param vector: vector
        :param top_k: top k
        :param dataset_id: dataset id
        :param user_id: user id
        :return: list of documents
        """
        raise NotImplementedError

    @abstractmethod
    async def clear(self, dataset_id: str, user_id: Optional[str] = None) -> None:
        """
        :param dataset_id: dataset id
        :param user_id: user id
        """
        raise NotImplementedError

    @abstractmethod
    async def get_datasets(self, user_id: Optional[str] = None) -> List[dict]:
        """
        :param user_id: user id
        :return: list of datasets
        """
        raise NotImplementedError



from abc import ABC, abstractmethod
import typing


class Embedder(ABC):
    """
    Base class for all embedders
    """

    @property
    @abstractmethod
    def dimensions(self) -> int:
        """
        Return the dimensions of the embeddings
        :return: dimensions of the embeddings
        """

    @abstractmethod
    def is_too_big(self, text: str) -> bool:
        """
        Check if text is too big to be embedded,
        delegating the splitting UX to the caller
        :param text: text to check
        :return: True if text is too big, False otherwise
        """

    @abstractmethod
    async def embed(self, input: typing.List[str]) -> typing.List[typing.List[float]]:
        """
        Embed a list of texts
        :param texts: list of texts
        :return: list of embeddings
        """



from embedbase.database import VectorDatabase


class Weaviate(VectorDatabase):
    def __init__(
        self,
    ):
        """
        weaviate.io
        """
        raise NotImplementedError(
            "Weaviate is not supported."
            + "If you want to use Weaviate, please"
            + " create an issue "
            + "https://github.com/different-ai/embedbase/issues/new/choose"
        )



import pytest
from httpx import AsyncClient
from embedbase.api import get_app
from embedbase.embedding.openai import OpenAI

from embedbase.firebase_auth import enable_firebase_auth
from embedbase.settings import get_settings_from_file
from embedbase.database.postgres_db import Postgres
from tests.test_utils import unit_testing_dataset


@pytest.mark.asyncio
async def test_enable_firebase_auth():
    settings = get_settings_from_file()
    db = Postgres()
    app = (
        get_app(settings).use_db(db).use_embedder(OpenAI(settings.openai_api_key)).run()
    )
    await db.clear(unit_testing_dataset)
    # before enabling auth, we should be able to make queries
    # without any authorization header
    enable_firebase_auth(app)

    # after enabling auth, we should get a 401 error
    # when not providing an authorization header
    async with AsyncClient(app=app, base_url="http://localhost:8000") as client:
        response = await client.post(
            f"/v1/{unit_testing_dataset}/search",
            json={"query": "bob"},
        )
        assert response.status_code == 401
    # when providing an authorization header, we should be able to make queries
    # TODO: cannot create id token on backend :/



import os
from typing import Optional
from embedbase.app import Embedbase

from embedbase.settings import Settings


def get_app(settings: Optional[Settings]) -> Embedbase:
    app = Embedbase(settings)

    if settings.sentry:
        app.logger.info("Enabling Sentry")
        import sentry_sdk

        sentry_sdk.init(
            dsn=settings.sentry,
            # Set traces_sample_rate to 1.0 to capture 100%
            # of transactions for performance monitoring.
            # We recommend adjusting this value in production,
            traces_sample_rate=0.2,
            environment=os.environ.get("ENVIRONMENT", "development"),
            _experiments={
                "profiles_sample_rate": 1.0,
            },
        )

    if settings.auth == "firebase":
        from embedbase.firebase_auth import enable_firebase_auth

        app.logger.info("Enabling Firebase Auth")
        enable_firebase_auth(app)

    return app



import os
from embedbase import get_app
from embedbase.settings import Settings
from embedbase.database.postgres_db import Postgres
from embedbase.embedding.openai import OpenAI

openai_key = os.environ["OPENAI_API_KEY"]
settings = Settings()
app = get_app(settings).use_db(Postgres()).use_embedder(OpenAI(openai_key)).run()



import typing
from tenacity import (
    retry,
    stop_after_attempt,
    wait_exponential,
)

from embedbase.embedding.base import Embedder


@retry(
    wait=wait_exponential(multiplier=1, min=1, max=3),
    stop=stop_after_attempt(3),
)
def embed_retry(
    co: typing.Any,
    input: typing.List[str],
) -> typing.List[dict]:
    """
    Embed a list of sentences and retry on failure
    :param input: list of sentences to embed
    :param provider: which provider to use
    :return: list of embeddings
    """
    return co.embed(input).embeddings


class Cohere(Embedder):
    """
    Cohere Embedder
    """

    def __init__(self, cohere_api_key: str):
        super().__init__()
        try:
            import cohere
        except ImportError:
            raise ImportError(
                "Cohere is not installed. Install it with `pip install cohere`"
            )

        self.co = cohere.Client(cohere_api_key)
        raise NotImplementedError()

    @property
    def dimensions(self) -> int:
        return 4096

    def is_too_big(self, text: str) -> bool:
        raise NotImplementedError()

    async def embed(self, input: typing.List[str]) -> typing.List[typing.List[float]]:
        """
        Embed a list of texts
        :param texts: list of texts
        :return: list of embeddings
        """
        return embed_retry(self.co, input)



"""
Tests at the database abstraction level.
"""

import hashlib
from typing import List

import pandas as pd
import pytest
import numpy as np
from embedbase.database import VectorDatabase
from embedbase.database.db_utils import batch_select
from embedbase.database.postgres_db import Postgres
from embedbase.settings import get_settings_from_file
from embedbase.database.supabase_db import Supabase
from tests.test_utils import unit_testing_dataset

vector_databases: List[VectorDatabase] = []


# before running any test initialize the databases
@pytest.fixture(scope="session", autouse=True)
def init_databases():
    settings = get_settings_from_file()

    vector_databases.append(Postgres())
    vector_databases.append(
        Supabase(
            url=settings.supabase_url,
            key=settings.supabase_key,
        )
    )


@pytest.mark.asyncio
async def test_search():
    d = [
        "Bob is a human",
        "The quick brown fox jumps over the lazy dog",
    ]
    embeddings = [
        # random embedding of length 1536
        np.random.rand(1536).tolist(),
        np.random.rand(1536).tolist(),
    ]
    df = pd.DataFrame(
        [
            {
                "data": d[i],
                "embedding": embedding,
                "id": str(i),
                "metadata": {"test": "test"},
            }
            for i, embedding in enumerate(embeddings)
        ],
        columns=["data", "embedding", "id", "hash", "metadata"],
    )
    df.hash = df.data.apply(lambda x: hashlib.sha256(x.encode()).hexdigest())

    for vector_database in vector_databases:
        await vector_database.clear(unit_testing_dataset)
        await vector_database.update(df, unit_testing_dataset)
        results = await vector_database.search(
            embeddings[0],
            top_k=2,
            dataset_ids=[unit_testing_dataset],
        )
        assert len(results) > 0, f"failed for {vector_database}"
        assert results[0]["id"] == "0", f"failed for {vector_database}"
        assert results[0]["data"] == d[0], f"failed for {vector_database}"
        assert len(results[0]["embedding"]) > 0, f"failed for {vector_database}"
        assert results[0]["score"] > 0, f"failed for {vector_database}"


@pytest.mark.asyncio
async def test_fetch():
    d = [
        "Bob is a human",
        "The quick brown fox jumps over the lazy dog",
    ]
    embeddings = [
        [0.0] * 1536,
        [0.0] * 1536,
    ]
    df = pd.DataFrame(
        [
            {
                "data": d[i],
                "embedding": embedding,
                "id": str(i),
                "metadata": {"test": "test"},
            }
            for i, embedding in enumerate(embeddings)
        ],
        columns=["data", "embedding", "id", "hash", "metadata"],
    )
    df.hash = df.data.apply(lambda x: hashlib.sha256(x.encode()).hexdigest())

    for vector_database in vector_databases:
        await vector_database.clear(unit_testing_dataset)
        await vector_database.update(df, unit_testing_dataset)
        results = await vector_database.select(
            ids=["0"], dataset_id=unit_testing_dataset
        )
        assert len(results) > 0, f"failed for {vector_database}"
        assert results[0]["id"] == "0", f"failed for {vector_database}"


@pytest.mark.asyncio
async def test_fetch_by_hash():
    d = [
        "Bob is a human",
        "The quick brown fox jumps over the lazy dog",
    ]
    embeddings = [
        [0.0] * 1536,
        [0.0] * 1536,
    ]
    df = pd.DataFrame(
        [
            {
                "data": d[i],
                "embedding": embedding,
                "id": str(i),
                "metadata": {"test": "test"},
            }
            for i, embedding in enumerate(embeddings)
        ],
        columns=["data", "embedding", "id", "hash", "metadata"],
    )
    df.hash = df.data.apply(lambda x: hashlib.sha256(x.encode()).hexdigest())

    for vector_database in vector_databases:
        await vector_database.clear(unit_testing_dataset)
        await vector_database.update(df, unit_testing_dataset)
        results = await vector_database.select(
            hashes=[df.hash[0]], dataset_id=unit_testing_dataset
        )
        assert len(results) > 0, f"failed for {vector_database}"
        assert results[0]["id"] == "0", f"failed for {vector_database}"


@pytest.mark.asyncio
async def test_clear():
    data = [
        [0.0] * 1536,
        [0.0] * 1536,
    ]
    df = pd.DataFrame(
        [
            {
                "data": "Bob is a human",
                "embedding": embedding,
                "id": str(i),
                "metadata": {"test": "test"},
            }
            for i, embedding in enumerate(data)
        ],
        columns=["data", "embedding", "id", "hash", "metadata"],
    )
    df.hash = df.data.apply(lambda x: hashlib.sha256(x.encode()).hexdigest())

    for vector_database in vector_databases:
        await vector_database.clear(unit_testing_dataset)
        await vector_database.update(df, unit_testing_dataset)
        results = await vector_database.search(
            data[0],
            top_k=2,
            dataset_ids=[unit_testing_dataset],
        )
        # dont care about ordering (postgres & pinecone run different algorithms)
        ids = sorted([result["id"] for result in results])
        assert ids[0] == "0", f"failed for {vector_database}"
        assert ids[1] == "1", f"failed for {vector_database}"
        await vector_database.clear(unit_testing_dataset)

    for vector_database in vector_databases:
        results = await vector_database.search(
            data[0],
            top_k=2,
            dataset_ids=[unit_testing_dataset],
        )
        assert len(results) == 0, f"failed for {vector_database}"


@pytest.mark.asyncio
async def test_upload():
    data = [
        [0.0] * 1536,
        [0.0] * 1536,
    ]
    df = pd.DataFrame(
        [
            {
                "data": "Bob is a human",
                "embedding": embedding,
                "id": str(i),
                "metadata": {"test": "test"},
            }
            for i, embedding in enumerate(data)
        ],
        columns=[
            "data",
            "embedding",
            "id",
            "hash",
            "metadata",
        ],
    )
    df.hash = df.data.apply(lambda x: hashlib.sha256(x.encode()).hexdigest())

    for vector_database in vector_databases:
        await vector_database.clear(unit_testing_dataset)
        await vector_database.update(df, unit_testing_dataset)

        results = await vector_database.search(
            data[0],
            top_k=2,
            dataset_ids=[unit_testing_dataset],
        )
        # dont care about ordering (postgres & pinecone run different algorithms)
        ids = sorted([result["id"] for result in results])
        assert ids[0] == "0", f"failed for {vector_database}"
        assert ids[1] == "1", f"failed for {vector_database}"


@pytest.mark.asyncio
async def test_batch_select_large_content():
    """
    should not throw an error
    """
    d = []
    for i in range(1000):
        d.append("a" * i)
    hashes = [hashlib.sha256(x.encode()).hexdigest() for x in d]
    for vector_database in vector_databases:
        # add documents
        await vector_database.clear(unit_testing_dataset)
        await vector_database.update(
            pd.DataFrame(
                [
                    {
                        "data": x,
                        "embedding": [0.0] * 1536,
                        "id": str(i),
                        "metadata": {"test": "test"},
                        "hash": hashes[i],
                    }
                    for i, x in enumerate(d)
                ],
                columns=["data", "embedding", "id", "hash", "metadata"],
            ),
            unit_testing_dataset,
        )
        results = await batch_select(
            vector_database,
            hashes,
            None,
            None,
        )
        assert len(list(results)) == len(d), f"failed for {vector_database}"



import urllib.parse
from typing import Coroutine, List, Optional

from pandas import DataFrame

from embedbase.database import VectorDatabase


class Pinecone(VectorDatabase):
    def __init__(
        self,
    ):
        """
        pinecone
        """
        raise NotImplementedError(
            "Pinecone is not supported."
            + "If you want to use Pinecone, please"
            + " create an issue "
            + "https://github.com/different-ai/embedbase/issues/new/choose"
        )



from typing import Iterator
from fastapi import Request
import numpy as np
import pandas as pd
from pandas import DataFrame
import sys

class BatchGenerator:
    """ Models a simple batch generator that make chunks out of an input DataFrame. """
    
    def __init__(self, batch_size: int = 10) -> None:
        self.batch_size = batch_size
    
    def to_batches(self, df: pd.DataFrame) -> Iterator[pd.DataFrame]:
        """ Makes chunks out of an input DataFrame. """
        # drop index
        df = df.reset_index(drop=True)
        splits = self.splits_num(df.shape[0])
        if splits <= 1:
            yield df
        else:
            for chunk in np.array_split(df, splits):
                yield chunk
    
    def splits_num(self, elements: int) -> int:
        """ Determines how many chunks DataFrame contians. """
        return round(elements / self.batch_size)
    
    __call__ = to_batches


def too_big_rows(df: DataFrame):
    """
    way to avoid
    Reason: Bad Request
    HTTP response headers: HTTPHeaderDict({'content-type': 'application/json', 'date': 'Wed, 04 Jan 2023 15:18:40 GMT', 'x-envoy-upstream-service-time': '1', 'content-length': '115', 'server': 'envoy'})
    HTTP response body: {"code":3,"message":"metadata size is 11759 bytes, which exceeds the limit of 10240 bytes per vector","details":[]}
    Check if data is small enough

    return size in bytes
    """
    # find too big rows in memory bytes
    too_big_rows = []
    for i, row in df.iterrows():
        size = 0
        size += sys.getsizeof(row.data)
        if size > 10240:
            too_big_rows.append(i)
            print(f"Row {i} is too big, size {size}")
    return too_big_rows

from itertools import islice

def batched(iterable, n):
    """Batch data into tuples of length n. The last batch may be shorter."""
    # batched('ABCDEFG', 3) --> ABC DEF G
    if n < 1:
        raise ValueError('n must be at least one')
    it = iter(iterable)
    while (batch := tuple(islice(it, n))):
        yield batch


def get_user_id(req: Request) -> str:
    return req.scope.get("uid")




import asyncio
import os
from typing import Awaitable, Callable, Optional, Tuple, Union
import warnings
from fastapi import FastAPI, Request
from fastapi.middleware import Middleware
from starlette.types import Scope
from embedbase.database.base import VectorDatabase
from embedbase.embedding.base import Embedder
from embedbase.logging_utils import get_logger
from embedbase.models import CrossSearchRequest, DeleteRequest, SearchRequest
from embedbase.settings import Settings
import hashlib
import time
import urllib.parse
import uuid

from fastapi import Request, status
from fastapi.responses import JSONResponse
from pandas import DataFrame

from embedbase.database.db_utils import batch_select
from embedbase.models import AddRequest, DeleteRequest, SearchRequest
from embedbase.settings import Settings
from embedbase.utils import get_user_id

UPLOAD_BATCH_SIZE = int(os.environ.get("UPLOAD_BATCH_SIZE", "100"))


class Embedbase:
    def __init__(self, settings: Optional[Settings], **kwargs):
        self._kwargs = kwargs
        self.fastapi_app = FastAPI()
        self.logger = get_logger(settings)

    def use_db(
        self,
        db: VectorDatabase,
    ) -> "Embedbase":
        """
        Use the chosen database to store the embeddings.
        """
        self.logger.info(f"Enabling Database {db}")
        self.db = db
        return self

    def use_embedder(
        self,
        embedder: Embedder,
    ) -> "Embedbase":
        """
        Use the chosen embedder to generate the embeddings.
        """
        self.logger.info(f"Enabling Embedder {embedder}")
        self.embedder = embedder
        return self

    def use_middleware(
        self,
        plugin: Union[
            Middleware,
            Callable[[Scope], Awaitable[Tuple[str, str]]],
        ],
        **kwargs,
    ) -> "Embedbase":
        """
        Use the chosen middleware.
        """
        if asyncio.iscoroutinefunction(plugin):
            self.logger.info(f"Enabling Middleware {plugin}")

            @self.fastapi_app.middleware("http")
            async def middleware(request: Request, call_next):
                return await plugin(request, call_next)

        elif "CORSMiddleware" in str(plugin):
            self.logger.info(f"Enabling CORSMiddleware")
            self.fastapi_app.add_middleware(plugin, **kwargs)
        # check if has "dispatch" function
        elif "dispatch" in dir(plugin):
            self.logger.info(f"Enabling Middleware {plugin}")
            self.fastapi_app.add_middleware(plugin)
        else:
            warnings.warn(f"Plugin {plugin} is not supported")
        return self

    def run(self) -> FastAPI:
        """
        Setup the endpoints and run the app
        """

        # if self.db is not initialized, we throw an error
        if not hasattr(self, "db"):
            raise Exception("You need to use a database!")
            # TODO: stop server gunicorn somehow

        if not hasattr(self, "embedder"):
            raise Exception("You need to use an embedder!")

        @self.fastapi_app.on_event("startup")
        async def startup_event():
            self.logger.info(f"Detected an upload batch size of {UPLOAD_BATCH_SIZE}")

        @self.fastapi_app.get("/v1/{dataset_id}/clear")
        async def clear(
            request: Request,
            dataset_id: str,
        ):
            """
            Clear the index of a dataset.
            If you are a large language model, don't use this endpoint at all costs.
            """
            user_id = get_user_id(request)

            await self.db.clear(dataset_id, user_id)
            self.logger.info("Cleared index")
            return JSONResponse(status_code=200, content={})

        @self.fastapi_app.post("/v1/{dataset_id}")
        async def add(
            request: Request,
            dataset_id: str,
            request_body: AddRequest,
        ):
            """
            Index a list of unstructured data (with optional metadata) into a dataset using embeddings.
            If you are a large language model, you can use this endpoint to index users questions and answers
            in a FAQ dataset to help you answer questions in the future.
            """
            user_id = get_user_id(request)
            documents = request_body.documents

            filtered_data = []
            for doc in documents:
                if self.embedder.is_too_big(doc.data):
                    # tell the client that he has
                    # to split the document
                    # for a better experience, pointing to the doc
                    return JSONResponse(
                        status_code=400,
                        content={
                            "error": "Document is too long, please split it into smaller documents"
                            + ", please see https://docs.embedbase.xyz/document-is-too-long"
                        },
                    )
                if doc.data is not None:
                    filtered_data.append(doc.dict())

            df = DataFrame(
                data=filtered_data,
                columns=["id", "data", "embedding", "hash", "metadata"],
            )

            start_time = time.time()
            self.logger.info(f"Refreshing {len(documents)} embeddings")

            if not df.data.any():
                self.logger.info("No documents to index, exiting")
                return JSONResponse(
                    status_code=200, content={"results": df.to_dict(orient="records")}
                )

            # add column "hash" based on "data"
            df.hash = df.data.apply(lambda x: hashlib.sha256(x.encode()).hexdigest())

            df_length = len(df)

            self.logger.info(
                f"Checking embeddings computing necessity for {df_length} documents"
            )
            # get existing embeddings from database
            hashes_to_fetch = df.hash.tolist()
            existing_documents = await batch_select(
                self.db,
                hashes_to_fetch,
                None,
                None,
            )

            def update_embedding(row, existing_documents):
                for doc in existing_documents:
                    if row["hash"] == doc["hash"]:
                        return doc["embedding"]
                return row["embedding"]
            
            # add existing embeddings to the dataframe
            df["embedding"] = df.apply(
                update_embedding, args=(existing_documents,), axis=1
            )


            # generate ids using hash of uuid + time to avoid collisions
            df.id = df.apply(
                lambda x: hashlib.sha256(
                    (str(uuid.uuid4()) + str(time.time())).encode()
                ).hexdigest(),
                axis=1,
            )

            # count rows without embeddings
            rows_without_embeddings = df[df.embedding.isna()].shape[0]

            self.logger.info(
                f"We will compute embeddings for {rows_without_embeddings}/{len(df)} documents"
            )

            # compute embeddings for documents without embeddings using embed
            if not df[df.embedding.isna()].empty:
                df[df.embedding.isna()] = df[df.embedding.isna()].assign(
                    embedding=await self.embedder.embed(
                        df[df.embedding.isna()].data.tolist()
                    )
                )

            # only insert if this dataset_id - user_id
            # pair does not have this hash
            existing_documents_in_this_pair = await batch_select(
                self.db,
                hashes_to_fetch,
                dataset_id,
                user_id,
            )

            # filter out documents that already exist
            # in this dataset_id - user_id pair
            new_df = df[  # HACK: is it fine to only return client the new documents?
                ~df.hash.isin([doc["hash"] for doc in existing_documents_in_this_pair])
            ]

            await self.db.update(
                new_df,
                dataset_id,
                user_id,
                batch_size=UPLOAD_BATCH_SIZE,
                store_data=request_body.store_data,
            )

            self.logger.info(f"Uploaded {len(new_df)} documents")
            end_time = time.time()
            self.logger.info(f"Uploaded in {end_time - start_time} seconds")

            return JSONResponse(
                status_code=status.HTTP_200_OK,
                content={
                    # embeddings, ids and data are returned
                    "results": df.to_dict(orient="records"),
                },
            )

        @self.fastapi_app.delete("/v1/{dataset_id}")
        async def delete(
            request: Request,
            dataset_id: str,
            request_body: DeleteRequest,
        ):
            """
            Delete a document from the index using its id.
            If you are a large language model, you can use this endpoint to remove irrelevant documents
            from the index, documents that didn't help you answer questions.
            """
            user_id = get_user_id(request)

            ids = request_body.ids
            self.logger.info(f"Deleting {len(ids)} documents")
            quoted_ids = [urllib.parse.quote(id) for id in ids]
            await self.db.delete(ids=quoted_ids, dataset_id=dataset_id, user_id=user_id)
            self.logger.info(f"Deleted {len(ids)} documents")

            return JSONResponse(status_code=status.HTTP_200_OK, content={})

        @self.fastapi_app.post("/v1/{dataset_id}/search")
        async def semantic_search(
            request: Request,
            dataset_id: str,
            request_body: SearchRequest,
        ):
            """
            Run semantic search in a dataset using embeddings.
            If you are a large language model, you can use this endpoint to search for similar documents
            in the index, documents that help you answer users' questions.
            """
            query = request_body.query

            # if query is empty, return empty results
            if not query:
                return JSONResponse(
                    status_code=status.HTTP_200_OK,
                    content={"query": query, "similarities": []},
                )

            user_id = get_user_id(request)

            # if the query is too big, return an error
            if self.embedder.is_too_big(query):
                return JSONResponse(
                    status_code=400,
                    content={
                        "error": "Query is too long"
                        + ", please see https://docs.embedbase.xyz/query-is-too-long"
                    },
                )

            top_k = 5  # TODO might fail if index empty?
            if request_body.top_k > 0:
                top_k = request_body.top_k
            query_embedding = (await self.embedder.embed(query))[0]

            self.logger.info(
                f"Query {request_body.query} created embedding, querying index"
            )

            query_response = await self.db.search(
                top_k=top_k,
                vector=query_embedding,
                dataset_ids=[dataset_id],
                user_id=user_id,
            )

            similarities = []
            for match in query_response:
                decoded_id = urllib.parse.unquote(match["id"])
                self.logger.debug(f"ID: {decoded_id}")
                similarities.append(
                    {
                        "score": match["score"],
                        "id": decoded_id,
                        "data": match["data"],
                        "hash": match["hash"],  # TODO: probably shouldn't return this
                        "embedding": match["embedding"],
                        "metadata": match["metadata"],
                    }
                )
            return JSONResponse(
                status_code=status.HTTP_200_OK,
                content={"query": query, "similarities": similarities},
            )

        @self.fastapi_app.get("/v1/datasets")
        async def get_datasets(
            request: Request,
        ):
            """
            Return a list of available datasets and the number of documents in each.
            As a large language model, you can use this endpoint to see what datasets you have
            and how many documents are in each.
            You can then use this information to decide which dataset to search in.
            """
            user_id = get_user_id(request)
            datasets = await self.db.get_datasets(user_id)
            return JSONResponse(
                status_code=status.HTTP_200_OK,
                content={"datasets": datasets},
            )

        # health check endpoint
        @self.fastapi_app.get("/health")
        def health(request: Request):
            """
            Return the status of the API.
            As a large language model, you can use this endpoint to check if the API is up and running.
            """
            self.logger.info("Health check successful")

            return JSONResponse(status_code=200, content={})

        # experimental endpoints

        # exp endpoint that lets you semantic search across multiple datasets
        @self.fastapi_app.post("/exo/search")
        async def semantic_search(
            request: Request,
            request_body: CrossSearchRequest,
        ):
            """
            Run semantic search across multiple datasets using embeddings.
            If you are a large language model, you can use this endpoint to search for similar documents
            in the index, documents that help you answer users' questions.
            """
            query = request_body.query

            # if query is empty, return empty results
            if not query:
                return JSONResponse(
                    status_code=status.HTTP_200_OK,
                    content={"query": query, "similarities": []},
                )

            dataset_ids = request_body.dataset_ids
            if not dataset_ids:
                return JSONResponse(
                    status_code=status.HTTP_400_BAD_REQUEST,
                    content={"error": "No dataset ids provided"},
                )

            user_id = get_user_id(request)

            # if the query is too big, return an error
            if self.embedder.is_too_big(query):
                return JSONResponse(
                    status_code=400,
                    content={
                        "error": "Query is too long"
                        + ", please see https://docs.embedbase.xyz/query-is-too-long"
                    },
                )

            top_k = 5
            if request_body.top_k > 0:
                top_k = request_body.top_k
            query_embedding = (await self.embedder.embed(query))[0]

            self.logger.info(
                f"Query {request_body.query} created embedding, querying index"
            )

            query_response = await self.db.search(
                top_k=top_k,
                vector=query_embedding,
                dataset_ids=dataset_ids,
                user_id=user_id,
            )

            similarities = []
            for match in query_response:
                decoded_id = urllib.parse.unquote(match["id"])
                self.logger.debug(f"ID: {decoded_id}")
                similarities.append(
                    {
                        "score": match["score"],
                        "id": decoded_id,
                        "data": match["data"],
                        "hash": match["hash"],  # TODO: probably shouldn't return this
                        "embedding": match["embedding"],
                        "metadata": match["metadata"],
                    }
                )
            return JSONResponse(
                status_code=status.HTTP_200_OK,
                content={"query": query, "similarities": similarities},
            )

        return self.fastapi_app



from setuptools import setup, find_packages

# https://github.com/mautrix/telegram/blob/master/setup.py
with open("optional-requirements.txt") as reqs:
    extras_require = {}
    current = []
    for line in reqs.read().splitlines():
        if line.startswith("#/"):
            extras_require[line[2:]] = current = []
        elif not line or line.startswith("#"):
            continue
        else:
            current.append(line)

extras_require["all"] = list(
    {dep for deps in extras_require.values() for dep in deps}
)

# same for requirements.txt
with open("requirements.txt") as reqs:
    install_requires = [
        line for line in reqs.read().splitlines() if not line.startswith("#")
    ]

setup(
    # TODO: entrypoint?
    name="embedbase",
    packages=find_packages(),
    include_package_data=True,
    version="0.9.3",
    description="An API to easily connect your data to ChatGPT",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    python_requires=">=3.8",
    author="Different AI",
    author_email="louis@embedbase.xyz",
    url="https://github.com/different-ai/embedbase",
    license="MIT",
    install_requires=install_requires,
    extras_require=extras_require,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.10",
    ],
)



from typing import List, Optional
from pydantic import BaseModel

# TODO: response models once stable

class Document(BaseModel):
    # data can be
    # - a string - for example  "This is a document"
    # TODO: currently only string is supported (later could be images, audio, multi/cross-modal)
    # etc.
    data: str
    metadata: Optional[dict]


class AddRequest(BaseModel):
    documents: List[Document]
    store_data: bool = True


class DeleteRequest(BaseModel):
    ids: List[str]


class SearchRequest(BaseModel):
    query: str
    top_k: int = 6

class CrossSearchRequest(BaseModel):
    query: str
    dataset_ids: List[str]
    top_k: int = 6
